@@ -1,6 +1,6 @@
 package Net::OpenSSH::Parallel;
 
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 
 use strict;
 use warnings;
@@ -156,13 +156,14 @@ sub push {
     my $selector = shift;
     my $action = shift;
     my $in_state = $self->{in_state};
-    my %opts = (ref $_[0] eq 'HASH' ? %{shift()} : ());
 
     if (ref $action eq 'CODE') {
 	$action = 'sub';
 	unshift @_, $action;
     }
 
+    my %opts = (($action ne 'sub' and ref $_[0] eq 'HASH') ? %{shift()} : ());
+
     my $alias = $action_alias{$action};
     $action = $alias if defined $alias;
 
@@ -171,13 +172,8 @@ sub push {
 
     my @labels = $self->_select_labels($selector);
 
-    if (keys %opts and
-	($action eq 'join' or
-	 $action eq 'sub')) {
-	croak "$action action does not accept options";
-    }
-
     if ($action eq 'join') {
+	keys %opts and croak "join action does not accept options";
 	my $notify_selector = shift @_;
 	my $join = { id => '#' . $self->{join_seq}++,
 		     depends => {},
@@ -191,8 +187,10 @@ sub push {
 
 	for my $label (@labels) {
 	    my $host = $self->{hosts}{$label};
-	    push @{$host->{queue}}, [$action, {}, $join];
+	    push @{$host->{queue}}, [join => {}, $join];
 	    $debug and _debug(api => "[$label] join $join->{id} queued");
+	    $self->_set_host_state($label, 'ready')
+		if $in_state->{done}{$label};
 	}
     }
     else {
@@ -285,7 +283,7 @@ sub _at_error {
     }
 
     delete $host->{current_task_reconns};
- 
+
     if ($on_error == OSSH_ON_ERROR_IGNORE) {
 	if ($error == OSSH_JOIN_FAILED) {
 	    $self->_set_host_state($label, 'ready');
@@ -445,6 +443,7 @@ sub _at_ready {
 	if ($action eq 'join') {
 	    my (undef, $join) = @$task;
 	    if ($join->{done}) {
+		$debug and _debug(join => "join[$join->{id}] is done");
 		if ($join->{failed}) {
 		    $self->_at_error($label, OSSH_JOIN_FAILED);
 		    return;
@@ -454,6 +453,7 @@ sub _at_ready {
 	    }
 	    CORE::push @{$join->{notify}}, $label;
 	    $self->_set_host_state($label, 'waiting');
+	    return;
 	}
 	elsif ($action eq '_notify') {
 	    my (undef, $join) = @$task;
@@ -854,7 +854,7 @@ base to some of its properties.
 
 For instance:
 
-  $pssh->queue('*', scp_get => "/var/log/messages", "messages.%HOST%");
+  $pssh->push('*', scp_get => "/var/log/messages", "messages.%HOST%");
 
 copies the log files appending the name of the remote hosts to the
 local file names.
@@ -1046,6 +1046,36 @@ hosts.
 
 Queues a call to a perl subroutine that will be executed locally.
 
+Note that subroutines are executed synchronously in the same process,
+so no other task will be scheduled while they run.
+
+=item join => $selector
+
+Joins allow to synchronize jobs between different servers.
+
+For instance:
+
+  $ssh->push('server_B', scp_get => '/tmp/foo', 'foo');
+  $ssh->push('server_A', join => 'server_B');
+  $ssh->push('server_A', scp_put => 'foo', '/tmp/foo');
+
+The join makes server_A to wait for the C<scp_get> operation queued in
+server_B to finish before proceeding with the C<scp_put>.
+
+In general the join will make the selected servers to wait for any
+task queued on the servers matched by C<$selector> to finish before
+proceeding with the next queued tasks.
+
+One common usage is to synchronize all servers at some point:
+
+  $ssh->push('*', join => '*');
+
+By default, errors are propagated at joins. For instance, in the
+example above, if the scp_get operation queued on server_B failed, it
+would abort any further operation queued on server_B and any further
+operation queued after the join in server_A. See also L</Error
+handling>.
+
 =back
 
 When given, C<%opts> can contain the following options:
@@ -1054,7 +1084,7 @@ When given, C<%opts> can contain the following options:
 
 =item on_error => $fail_mode
 
-=item on_error => sub { }
+=item on_error => sub { ... }
 
 See L</Error handling>.
 
@@ -1062,7 +1092,7 @@ See L</Error handling>.
 
 not implemented yet!
 
-=item on_done => sub { }
+=item on_done => sub { ... }
 
 not implemented yet!
 
@@ -1176,7 +1206,7 @@ several host in parallel.
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright E<copy> 2009-2010 by Salvador FandiE<ntilde>o
+Copyright E<copy> 2009, 2010 by Salvador FandiE<ntilde>o
 (sfandino@yahoo.com).
 
 This library is free software; you can redistribute it and/or modify
